from remin.solver import Solver, make_trainer
from remin.residual import Residual, make_loader
from remin import func, domain
import torch
from torch import nn
from model import Heat

torch.set_default_device('cuda')
torch.set_float32_matmul_precision('high')
    
model = Heat()

u = func.functional_call(model)
u_t = func.grad(u, argnum=1)
u_x = func.grad(u, argnum=2)
u_xx = func.grad(u_x, argnum=2)

def pde_residual(params, t, x):
    return u_t(params, t, x) - 0.05*u_xx(params, t, x)

def ic_residual(params, t, x):
    return u(params, t, x) - torch.tensor(1.0)

def bc_residual(params, t, x):
    return u(params, t, x) - torch.tensor(0.0)

pde_col = domain.HyperCube([(0, 1), (0, 1)], 1000)
ic_col = domain.Line((0, 0), (0, 1), 1000)
bc0_col = domain.Line((0, 0), (1, 0), 1000)
bc1_col = domain.Line((0, 1), (1, 1), 1000)

pde_res = Residual(pde_col, pde_residual)
ic_res = Residual(ic_col, ic_residual)
bc0_res = Residual(bc0_col, bc_residual)
bc1_res = Residual(bc1_col, bc_residual)

if __name__ == '__main__':
    loader = make_loader(
        [pde_res, ic_res, bc0_res, bc1_res],
        fully_loaded=True,
        #batched=False,
        #pin_memory=True,
        #num_workers=1
    )

    epochs = 35000
    lr = 2e-4
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    lossfunc = nn.MSELoss()
    trainer = make_trainer(model, loader, optimizer, lossfunc)
    heat = Solver('heat', 'outputs',trainer=trainer)

    heat.compile(backend='inductor', fullgraph=True)

    heat.fit(epochs)



import torch
from remin.residual import Loader
from remin.solver.residual_loss import ResidualLoss
from ..residual import Loader, FullLoader, BatchLoader
from .residual_loss import FuncLoss


class Trainer:

    def __init__(self,
                 loader: Loader,
                 model: torch.nn.Module = None,
                 optimizer: torch.optim.Optimizer = None,
                 residual_loss: ResidualLoss = None,
                 metric_loss: ResidualLoss = None) -> None:
        self.model = model
        self.device = None
        self.params = None

        self.loader = loader
        self.optimizer = optimizer
        self.residual_loss = residual_loss
        self.metric_loss = metric_loss

    def setup(self, model):
        self.model = model
        self.device = next(model.parameters()).device
        self.params = dict(model.named_parameters())
        self.residual_loss.setup(self.model, self.loader)
        if self.metric_loss:
            self.metric_loss.setup(self.model, self.loader)

    def __call__(self) -> float:
        raise NotImplementedError('Train Loop must be defined.')

    def eval_loss(self) -> float:
        raise NotImplementedError('Loss evaluation must be defined.')

    def optim_closure(self, xs):

        def closure():
            self.optimizer.zero_grad()
            resloss = self.residual_loss(self.params, xs)
            resloss.backward()
            return resloss

        return closure


class FullyLoadedTrainer(Trainer):

    def setup(self, model):
        super().setup(model)
        self.loader.instanciate(self.device)

    def __call__(self) -> float:
        resloss = self.residual_loss(self.params, self.loader.xs).item()
        self.optimizer.step(self.optim_closure(self.loader.xs))
        return resloss

    def eval_loss(self) -> float:
        self.model.eval()
        metloss = self.metric_loss(self.params, self.loader.xs).item()
        self.model.train()
        return metloss


class BatchedTrainer(Trainer):

    def __call__(self) -> float:
        resloss = 0.0
        for batch in self.loader.loader:
            xs = torch.vsplit(batch.to(self.device, non_blocking=True),
                              self.loader.vsplit)
            resloss += self.residual_loss(self.params, xs).item()
            self.optimizer.step(self.optim_closure(xs))
        resloss /= self.loader.batch_sampler.n_batches
        return resloss

    def eval_loss(self) -> float:
        metloss = 0.0
        self.model.eval()
        for batch in self.loader.loader:
            xs = torch.vsplit(batch.to(self.device, non_blocking=True),
                              self.loader.vsplit)
            metloss += self.metric_loss(self.params, xs).item()
        metloss /= self.loader.batch_sampler.n_batches
        self.model.train()
        return metloss


def make_trainer(loader: Loader,
                 model: torch.nn.Module = None,
                 optimizer: torch.optim.Optimizer = None,
                 lossfunc=None,
                 residual_loss=None,
                 metric_loss=None):
    if residual_loss is None:
        residual_loss = FuncLoss(lossfunc, model, loader)
    if isinstance(loader, BatchLoader):
        return BatchedTrainer(loader, model, optimizer, residual_loss, metric_loss)
    if isinstance(loader, FullLoader):
        return FullyLoadedTrainer(loader, model, optimizer, residual_loss, metric_loss)
    raise ValueError('Unknown Loader Type.')



import torch
from model import Heat
import matplotlib.pyplot as plt
from matplotlib import cm


model = Heat()
mdata = torch.load('outputs/heat_best_model.pt')
model.load_state_dict(mdata['model_state_dict'])
model.eval()

t = torch.linspace(0, 1, 100)
x = torch.linspace(0, 1, 100)
T, X = torch.meshgrid(t, x, indexing='xy')

u_pred = model(
    T.flatten()[:, None],
    X.flatten()[:, None]
    ).reshape_as(T)

fig, ax1 = plt.subplots(1,1,subplot_kw={'projection': '3d'})
fig.set_size_inches(6, 4)
fig.suptitle('Heat Equation Solution')

ax1.set(xlabel='x', ylabel='t', zlabel='u')

surf = ax1.plot_surface(
    X.detach(),
    T.detach(),
    u_pred.detach(),
    linewidth=0, antialiased=False, cmap=cm.coolwarm)

fig.savefig('outputs/heat.png', dpi=300)


import torch
from torch.utils.data import (Dataset, ConcatDataset, BatchSampler, DataLoader)


class Residual:

    def __init__(self, domain, equations, weight=1.0, batch_size=None):
        self.domain = domain
        self.weight = weight
        if isinstance(equations, (list, tuple)):
            self.equations = equations
        else:
            self.equations = [equations]
        if batch_size is None:
            self.batch_size = len(domain)
        else:
            self.batch_size = batch_size

    def to_torch(self, device):
        return torch.from_numpy(self.domain.astype('float32')).to(device)

    @staticmethod
    def split_domain(domain):
        if domain.shape[1] == 1:
            domain.requires_grad_()
            return (domain, )
        xs = torch.hsplit(domain, domain.shape[1])
        for x in xs:
            x.requires_grad_()
        return xs


class ResidualDataset(Dataset):

    def __init__(self, domain) -> None:
        self.domain = torch.from_numpy(domain.astype('float32'))

    def __len__(self):
        return len(self.domain)

    def __getitem__(self, index):
        return self.domain[index]


class ResBatchSampler(BatchSampler):

    def __init__(self, cumulative_sizes, batch_sizes) -> None:
        self.cumul_sizes = cumulative_sizes
        self.batch_sizes = batch_sizes
        self.batch_size = sum(batch_sizes)
        self.ndata = len(batch_sizes)
        self.n_batches = self.cumul_sizes[0] // self.batch_sizes[0]
        for i in range(1, self.ndata):
            n = (self.cumul_sizes[i] - self.cumul_sizes[i - 1]) // self.batch_sizes[i]
            if n > self.n_batches:
                self.n_batches = n

        self.batch = [None] * self.batch_size
        self.batch_shift = [None] * self.batch_size
        start_idx, batch_idx = 0, 0
        for i in range(self.ndata):
            batch_size = self.batch_sizes[i]
            cumul_size = self.cumul_sizes[i]
            prev_cumul_size = self.cumul_sizes[i - 1] if i > 0 else 0
            for j in range(batch_size):
                self.batch_shift[batch_idx + j] = (batch_size, cumul_size,
                                                   prev_cumul_size)
                self.batch[batch_idx + j] = start_idx + j
            start_idx = cumul_size
            batch_idx += batch_size

    def __iter__(self):
        for _ in range(self.n_batches):
            yield self.batch
            for j in range(self.batch_size):
                batch_size, cumul_size, prev_cumul_size = self.batch_shift[j]
                self.batch[j] = self.batch[j] + batch_size
                if self.batch[j] >= cumul_size:
                    self.batch[j] -= cumul_size - prev_cumul_size


class Loader:

    def __init__(self, residuals, **kwargs) -> None:
        self.residuals = residuals
        self.n_res = len(residuals)
        self.functions = [None] * self.n_res
        self.weights = [None] * self.n_res
        self.kwargs = kwargs

    def instanciate(self, device):
        raise NotImplementedError('Base class cannot be instanciated.')


class FullLoader(Loader):

    def __init__(self, residuals, **kwargs) -> None:
        super().__init__(residuals, **kwargs)
        self.xs = [None] * self.n_res

    def instanciate(self, device):
        for i in range(self.n_res):
            self.functions[i] = self.residuals[i].equations
            self.xs[i] = self.residuals[i].to_torch(device)
            self.weights[i] = self.residuals[i].weight


class BatchLoader(Loader):

    def __init__(self, residuals, batched, **kwargs) -> None:
        super().__init__(residuals, **kwargs)
        self.batch_sizes = [None] * self.n_res
        self.datasets = [None] * self.n_res
        self.batch_sampler = None
        self.batched = batched

        for i in range(self.n_res):
            self.batch_sizes[i] = self.residuals[i].batch_size
            self.functions[i] = self.residuals[i].equations
            self.weights[i] = self.residuals[i].weight
            self.datasets[i] = ResidualDataset(self.residuals[i].domain)

        self.dataset = ConcatDataset(self.datasets)
        self.dataset_size = len(self.dataset)
        self.vsplit = [self.batch_sizes[0]]

        for batch_size in self.batch_sizes[1:-1]:
            self.vsplit.append(self.vsplit[-1] + batch_size)
        if self.batched:
            self.batch_sampler = ResBatchSampler(self.dataset.cumulative_sizes,
                                                 self.batch_sizes)
            self.loader = DataLoader(self.dataset,
                                     batch_sampler=self.batch_sampler,
                                     **self.kwargs)
        else:
            self.loader = DataLoader(self.dataset,
                                     batch_size=self.dataset_size,
                                     **self.kwargs)

    def instanciate(self, device):
        pass


def make_loader(residuals, fully_loaded=False, batched=False, **kwargs):
    if fully_loaded and batched:
        raise ValueError('Data can not be fully loaded and batched at the same time.')
    if fully_loaded:
        return FullLoader(residuals, **kwargs)
    return BatchLoader(residuals, batched, **kwargs)



import numpy as np
from pyDOE.doe_lhs import lhs


def HyperCube(dimensions, n_coll):
    ndim = len(dimensions)
    scale = np.array([dim[1] - dim[0] for dim in dimensions])
    shift = np.array([dim[0] for dim in dimensions])
    return lhs(ndim, n_coll) * scale + shift


def Circle(center, radius, n_coll):
    center = np.array(center)
    coll = lhs(2, n_coll)
    radius, theta = coll[0], 2 * np.pi * coll[1]
    return radius * np.hstack((np.cos(theta), np.sin(theta))) + center


def Line(pt1, pt2, n_coll):
    ndim = len(pt1)
    scale = np.array([x2 - x1 for x1, x2 in zip(pt1, pt2)])
    shift = np.array(pt1)
    return lhs(ndim, n_coll) * scale + shift


def Point(pt):
    return np.array([pt])


def Ring(center, radius, n_coll):
    center = np.array(center)
    theta = 2 * np.pi * lhs(1, n_coll)
    return radius * np.hstack((np.cos(theta), np.sin(theta))) + center


def cut(domain_from, constraints):
    if not isinstance(constraints, list):
        constraints = [constraints]
    cond = np.full_like(domain_from[:, 0], True)
    for constraint in constraints:
        cond = np.logical_and(cond, constraint(domain_from))
    return domain_from[cond]


def union(*args):
    return np.vstack(args)


def shuffle(domain):
    np.random.shuffle(domain)
    return domain



import torch
from torch import nn

class Heat(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear_sigmoid_stack = nn.Sequential(
            nn.Linear(2, 32),
            nn.Tanh(),
            nn.Linear(32, 32),
            nn.Tanh(),
            nn.Linear(32, 32),
            nn.Tanh(),
            nn.Linear(32, 32),
            nn.Tanh(),
            nn.Linear(32, 1)
        )
    
    def forward(self, t, x):
        xs = torch.hstack((t, x))
        return self.linear_sigmoid_stack(xs)


from functools import wraps
from torch import func, autograd
from torch import _dynamo as dynamo


def make_sum(f):

    @wraps(f)
    def _sum(*args, **kwargs):
        return f(*args, **kwargs).sum()

    return _sum


def take_col_sum(f, i):

    @wraps(f)
    def _colsum(*args, **kwargs):
        return f(*args, **kwargs)[:, i].sum()

    return _colsum


def traceable(f):
    f = dynamo.allow_in_graph(f)

    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


def functional_call(model):

    def _call(params, *args):
        return func.functional_call(model, dict(params), args)

    return _call


def fgrad(f, argnum=1):
    return func.grad(make_sum(f), argnums=argnum)


def fgradi(f, i, argnum=1):
    return func.grad(take_col_sum(f, i), argnums=argnum)


def grad(u, xs, create_graph=True, retain_graph=True):
    return autograd.grad(u.sum(),
                         xs,
                         create_graph=create_graph,
                         retain_graph=retain_graph)



from time import thread_time
from typing import Dict, List, Any
import numpy as np
import matplotlib.pyplot as plt


class Callback:
    # Callback interface
    def __init__(self) -> None:
        self.state_dict: Dict[str, Any] = {}

    def on_train_begin(self):
        pass

    def on_train_end(self):
        pass

    def on_epoch_begin(self):
        pass

    def on_epoch_end(self):
        pass


class UnionCallback(Callback):

    def __init__(self, state_dict: Dict[str, Any], *callbacks: List[Callback]) -> None:
        super().__init__()
        self.callbacks = callbacks
        self.state_dict = state_dict
        for callback in callbacks:
            callback.state_dict = state_dict

    def on_train_begin(self):
        for callback in self.callbacks:
            callback.on_train_begin()

    def on_train_end(self):
        for callback in self.callbacks:
            callback.on_train_end()

    def on_epoch_begin(self):
        for callback in self.callbacks:
            callback.on_epoch_begin()

    def on_epoch_end(self):
        rval = 0
        for callback in self.callbacks:
            if callback.on_epoch_end():
                rval = -1
        return rval

    def reset(self, *callbacks):
        self.callbacks = callbacks
        for callback in callbacks:
            callback.state_dict = self.state_dict

    def append(self, *callbacks):
        self.callbacks += callbacks


class TotalTimeCallback(Callback):

    def __init__(self) -> None:
        super().__init__()
        self.t0 = 0

    def on_train_begin(self):
        self.t0 = thread_time()

    def on_train_end(self):
        print(f'Total training time: {((thread_time() - self.t0) / 60):7.3f}mins.')


class LogCallback(Callback):

    def __init__(self, log_epoch=100, log_progress=10) -> None:
        super().__init__()
        self.log_epoch = log_epoch
        self.log_progress = log_progress
        self.t0 = 0
        self.t_ave = 0
        self.metric = False

    def on_train_begin(self):
        self.metric = bool(self.state_dict['solver'].trainer.metric_loss)

    def on_epoch_begin(self):
        self.t0 = thread_time()

    def on_epoch_end(self):
        epoch = self.state_dict['epoch']
        epochs = self.state_dict['epochs']
        resloss = self.state_dict['resloss']
        metloss = self.state_dict['metloss']

        self.t_ave = (self.t_ave * epoch + (thread_time() - self.t0)) / (epoch + 1)
        if epoch % self.log_progress == 0:
            prefix = f'Loss: {resloss:10.6f}'
            if self.metric:
                prefix += f' - Metric Loss: {metloss:10.6f}'
            self._printProgress(epoch, epochs, self.log_epoch, prefix,
                                f'{(self.t_ave*1e3):7.3f}ms/epoch')

    @staticmethod
    def _printProgress(epoch,
                       epochs,
                       log_epoch,
                       prefix='',
                       suffix='',
                       length=20,
                       fill='#',
                       printEnd='\r'):
        filledLength = int(length * (epoch % log_epoch) // log_epoch)
        if epoch % log_epoch == 0:
            filledLength = length
            printEnd = '\n'
        loading_bar = fill * filledLength + '.' * (length - filledLength)
        print(f'{prefix} [{loading_bar}] [{epoch:>5d}/{epochs:>5d}] {suffix}',
              flush=True,
              end=printEnd)


class SaveCallback(Callback):

    def __init__(self, best_suffix='_best.pt', final_suffix='_final.pt') -> None:
        super().__init__()
        self.best_loss = float('inf')
        self.best_suffix = best_suffix
        self.final_suffix = final_suffix

    def on_epoch_end(self):
        resloss = self.state_dict['resloss']
        epoch = self.state_dict['epoch']
        solver = self.state_dict['solver']
        if resloss < self.best_loss:
            self.best_loss = resloss
            solver.save(epoch, self.best_suffix)

    def on_train_end(self):
        solver = self.state_dict['solver']
        epoch = self.state_dict['epoch']
        solver.save(epoch, self.final_suffix)


class PlotCallback(Callback):

    def __init__(self,
                 state='resloss',
                 size_inches=(10, 5),
                 color='red',
                 linestyle='-',
                 linewidth=2,
                 title='Residual Loss(L)',
                 ylabel='Loss',
                 xlabel='Epoch',
                 yscale='log',
                 grid_on=True,
                 name='L_vs_epoch.png',
                 dpi=300) -> None:
        super().__init__()
        self.state = state
        self.size_inches = size_inches
        self.color = color
        self.linestyle = linestyle
        self.linewidth = linewidth
        self.title = title
        self.ylabel = ylabel
        self.xlabel = xlabel
        self.yscale = yscale
        self.fig_name = name
        self.grid_on = grid_on
        self.dpi = dpi
        self.losses = None
        self.save_path = None

    def on_train_begin(self):
        self.losses = np.zeros((self.state_dict['epochs'], ))
        self.save_path = self.state_dict['solver'].save_path

    def on_epoch_end(self):
        self.losses[self.state_dict['epoch'] - 1] = self.state_dict[self.state]

    def on_train_end(self):
        fig, ax = plt.subplots(1, 1)
        if self.size_inches:
            fig.set_size_inches(*self.size_inches)
        ax.plot(self.losses,
                color=self.color,
                linestyle=self.linestyle,
                linewidth=self.linewidth)
        ax.set_yscale(self.yscale)
        ax.set(title=self.title, ylabel=self.ylabel, xlabel=self.xlabel)
        ax.grid(self.grid_on)
        plt.savefig(self.save_path + self.fig_name, dpi=self.dpi)



import torch
from ..residual import Residual, Loader


class ResidualLoss:

    def __init__(self,
                 lossfunc,
                 model: torch.nn.Module = None,
                 loader: Loader = None) -> None:
        self.loader = loader
        self.lossfunc = lossfunc
        self.model = model
        self.__name__ = 'residual_loss'

    def setup(self, model, loader):
        self.model = model
        self.loader = loader

    def __call__(self, params, xs) -> float:
        raise NotImplementedError('Loss must be defined.')


class FuncLoss(ResidualLoss):

    def __call__(self, params, xs) -> float:
        resloss = 0
        for i in range(self.loader.n_res):
            domain = Residual.split_domain(xs[i])
            for func in self.loader.functions[i]:
                resloss += self.loader.weights[i] * self.lossfunc(
                    func(params, *domain), torch.zeros_like(domain[0]))
        return resloss


class EagerLoss(ResidualLoss):

    def __call__(self, params, xs) -> float:
        resloss = 0
        for i in range(self.loader.n_res):
            domain = Residual.split_domain(xs[i])
            U = self.model(*domain)
            for func in self.loader.functions[i]:
                resloss += self.loader.weights[i] * self.lossfunc(
                    func(U, *domain), torch.zeros_like(domain[0]))
        return resloss


class ModLoss(ResidualLoss):

    def __call__(self, params, xs) -> float:
        resloss = 0
        for i in range(self.loader.n_res):
            domain = Residual.split_domain(xs[i])
            for func in self.loader.functions[i]:
                resloss += self.loader.weights[i] * self.lossfunc(
                    func(self.model, *domain), torch.zeros_like(domain[0]))
        return resloss



import os
from typing import List
import torch
from ..func import traceable
from ..callbacks import Callback, UnionCallback, SaveCallback, LogCallback
from .trainer import Trainer


class Solver:

    def __init__(self,
                 model: torch.nn.Module,
                 name: str = 'pinn',
                 save_folder: str = '.',
                 trainer: Trainer = None,
                 callbacks: List[Callback] = None) -> None:
        self.name = name
        self.save_path = save_folder + '/'

        self.trainer = trainer
        self.model = model
        self.trainer.setup(self.model)

        self.epoch = 1
        if callbacks is None:
            callbacks = [SaveCallback(), LogCallback()]

        self.state_dict = {
            'solver': self,
            'epoch': None,
            'epochs': None,
            'resloss': None,
            'metloss': None
        }
        self.callback = UnionCallback(self.state_dict, *callbacks)

    def load(self, filename):
        assert self.model is not None
        state_dict = torch.load(filename)
        self.model.load_state_dict(state_dict['model_state_dict'])
        self.trainer.optimizer.load_state_dict(state_dict['optimizer_state_dict'])
        self.trainer.residual_loss.lossfunc = state_dict['loss']

    def save(self, epoch, suffix=''):
        if not os.path.isdir(self.save_path):
            os.makedirs(self.save_path)
        torch.save(
            {
                'epoch': epoch + 1,
                'model_state_dict': self.model.state_dict(),
                'optimizer_state_dict': self.trainer.optimizer.state_dict(),
                'loss': self.trainer.residual_loss.lossfunc,
            }, self.save_path + self.name + suffix)

    def compile(self, backend='inductor', fullgraph=False):
        if self.trainer.residual_loss is None:
            raise ValueError('Residual Loss must be defined.')
        if self.trainer.residual_loss.lossfunc is None:
            raise ValueError('Loss function must be defined.')

        self.trainer.residual_loss = torch.compile(traceable(
            self.trainer.residual_loss),
                                                   backend=backend,
                                                   fullgraph=fullgraph)

    def update_state(self, **state):
        for key in state.keys():
            self.state_dict[key] = state[key]

    def append_callbacks(self, *callbacks):
        self.callback.append(*callbacks)

    def reset_callbacks(self, *callbacks):
        self.callback.reset(*callbacks)

    def fit(self, epochs=1):
        if self.trainer.residual_loss is None:
            raise ValueError('Residual Loss must be defined.')
        if self.trainer.residual_loss.lossfunc is None:
            raise ValueError('Loss function must be defined.')
        if self.trainer.optimizer is None:
            raise ValueError('Optimizer must be defined.')

        self.update_state(epochs=epochs)

        self.callback.on_train_begin()
        for epoch in range(1, epochs + 1):
            self.callback.on_epoch_begin()
            resloss = self.trainer()
            if self.trainer.metric_loss:
                metloss = self.trainer.eval_loss()
                self.update_state(metloss=metloss)
            self.update_state(resloss=resloss, epoch=epoch)
            if self.callback.on_epoch_end():
                break
        self.callback.on_train_end()
        self.epoch = epoch
